import logging
import sys

from .client import Subscriber
from .middleware import run_middleware_hook
from .subscription import Callback

logger = logging.getLogger(__name__)


class Worker:
    """A Worker manages the subscriptions which consume Google PubSub messages.

    Facilitates the creation of subscriptions if not already created,
    and the starting and stopping the consumption of them.

    :param subscriptions: list :class:`~rele.subscription.Subscription`
    """
    def __init__(self, subscriptions, gc_project_id, credentials):
        self._subscriber = Subscriber(gc_project_id, credentials)
        self._futures = []
        self._subscriptions = subscriptions

    def setup(self):
        """Create the subscriptions on a Google PubSub topic.

        If the subscription already exists, the subscription will not be
        re-created. Therefore, it is idempotent.
        """
        for subscription in self._subscriptions:
            self._subscriber.create_subscription(subscription.name,
                                                 subscription.topic)

    def start(self):
        """Begin consuming all subscriptions.

        When consuming a subscription, a `StreamingPullFuture` is returned from
        the Google PubSub client library. This future can be used to
        manage the background stream.

        The futures are stored so that they can be cancelled later on
        for a graceful shutdown of the worker.
        """
        for subscription in self._subscriptions:
            self._futures.append(self._subscriber.consume(
                subscription_name=subscription.name,
                callback=Callback(subscription)
            ))

    def stop(self, signal=None, frame=None):
        """Manage the shutdown process of the worker.

        This function has two purposes:

            1. Cancel all the futures created.
            2. And close all the database connections
               opened by Django. Even though we cancel the connections
               for every execution of the callback, we want to be sure
               that all the database connections are closed
               in this process.

        Exits with code 0 for a clean exit.

        :param signal: Needed for `signal.signal <https://docs.python.org/3/library/signal.html#signal.signal>`_  # noqa
        :param frame: Needed for `signal.signal <https://docs.python.org/3/library/signal.html#signal.signal>`_  # noqa
        """
        run_middleware_hook('pre_worker_stop', self._subscriptions)
        for future in self._futures:
            future.cancel()

        run_middleware_hook('post_worker_stop')
        sys.exit(0)



import logging
import signal
import time

from django.apps import apps
from django.conf import settings
from django.core.management import BaseCommand
from django.utils.module_loading import module_has_submodule

from rele import Worker
import rele

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Start subscriber threads to consume Relé topics.'

    def handle(self, *args, **options):
        subs = self._autodiscover_subs()
        self.stdout.write(f'Configuring worker with {len(subs)} '
                          f'subscription(s)...')
        for sub in subs:
            self.stdout.write(f'  {sub}')
        worker = Worker(subs,
                        settings.RELE['GC_PROJECT_ID'],
                        settings.RELE['GC_CREDENTIALS'])
        worker.setup()
        worker.start()

        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, worker.stop)
        signal.signal(signal.SIGTSTP, worker.stop)

        self._wait_forever()

    def _discover_subs_modules(self):
        logger.debug('Autodiscovering subs...')
        app_configs = apps.get_app_configs()
        subs_modules = []
        for conf in app_configs:
            if module_has_submodule(conf.module, "subs"):
                module = conf.name + ".subs"
                subs_modules.append(module)
                self.stdout.write(" * Discovered subs module: %r" % module)
        return subs_modules

    def _autodiscover_subs(self):
        return rele.config.load_subscriptions_from_paths(
            self._discover_subs_modules(),
            settings.RELE['SUB_PREFIX'],
            settings.RELE.get('FILTER_SUBS_BY'))

    def _wait_forever(self):
        self.stdout.write('Consuming subscriptions...')
        while True:
            time.sleep(1)



import importlib

_middlewares = []


default_middleware = ['rele.contrib.LoggingMiddleware']


def register_middleware(config):
    paths = config.middleware
    global _middlewares
    _middlewares = []
    for path in paths:
        *module_parts, middleware_class = path.split('.')
        module_path = '.'.join(module_parts)
        module = importlib.import_module(module_path)
        middleware = getattr(module, middleware_class)()
        middleware.setup(config)
        _middlewares.append(middleware)


def run_middleware_hook(hook_name, *args, **kwargs):
    for middleware in _middlewares:
        getattr(middleware, hook_name)(*args, **kwargs)


class BaseMiddleware:

    def setup(self, config):
        pass

    def pre_publish(self, topic, data, attrs):
        pass

    def post_publish(self, topic):
        pass

    def pre_process_message(self, subscription):
        pass

    def post_process_message(self):
        pass

    def post_process_message_success(self, subscription, start_time):
        pass

    def post_process_message_failure(
            self, subscription, exception, start_time):
        pass

    def pre_worker_start(self):
        pass

    def post_worker_start(self):
        pass

    def pre_worker_stop(self, subscriptions):
        pass

    def post_worker_stop(self):
        pass



import logging
import time

from rele.middleware import BaseMiddleware


class LoggingMiddleware(BaseMiddleware):

    def __init__(self):
        self._logger = None

    def setup(self, config):
        self._logger = logging.getLogger(__name__)
        self._app_name = config.app_name

    def _build_data_metrics(
            self, subscription, status, start_processing_time=None):
        result = {
            'agent': self._app_name,
            'topic': subscription.topic,
            'status': status,
            'subscription': subscription.name,
        }

        if start_processing_time is not None:
            end_processing_time = time.time()
            result['duration_seconds'] = round(
                end_processing_time - start_processing_time, 3)

        return result

    def pre_publish(self, topic, data, attrs):
        self._logger.info(
            f'Publishing to {topic}',
            extra={
                'pubsub_publisher_attrs': attrs,
                'metrics': {
                    'name': 'publications',
                    'data': {
                        'agent': self._app_name,
                        'topic': topic,
                    }
                }
            })

    def pre_process_message(self, subscription):
        self._logger.debug(
            f'Start processing message for {subscription}',
            extra={
                'metrics': {
                    'name': 'subscriptions',
                    'data': self._build_data_metrics(subscription, 'received')
                }
            })

    def post_process_message_success(self, subscription, start_time):
        self._logger.info(
            f'Successfully processed message for {subscription}',
            extra={
                'metrics': {
                    'name': 'subscriptions',
                    'data': self._build_data_metrics(
                        subscription, 'succeeded', start_time)
                }
            })

    def post_process_message_failure(
            self, subscription, exception, start_time):
        self._logger.error(
            f'Exception raised while processing message '
            f'for {subscription}: {str(exception.__class__.__name__)}',
            exc_info=True,
            extra={
                'metrics': {
                    'name': 'subscriptions',
                    'data': self._build_data_metrics(
                        subscription, 'failed', start_time)
                }
            })

    def pre_worker_stop(self, subscriptions):
        self._logger.info(
            f'Cleaning up {len(subscriptions)} subscription(s)...')



import json
import logging
import time

from .middleware import run_middleware_hook

logger = logging.getLogger(__name__)


class Subscription:
    """The Subscription class

    """
    def __init__(self, func, topic, prefix='', suffix='', filter_by=None):
        self._func = func
        self.topic = topic
        self._prefix = prefix
        self._suffix = suffix
        self._filter_by = filter_by

    @property
    def name(self):
        name_parts = [self._prefix, self.topic, self._suffix]
        return '-'.join(filter(lambda x: x, name_parts))

    @property
    def prefix(self):
        return self._prefix

    def set_prefix(self, prefix):
        self._prefix = prefix

    @property
    def filter_by(self):
        return self._filter_by

    def set_filter_by(self, filter_by):
        self._filter_by = filter_by

    def __call__(self, data, **kwargs):
        if 'published_at' in kwargs:
            kwargs['published_at'] = float(kwargs['published_at'])

        if self._filter_returns_false(kwargs):
            return

        return self._func(data, **kwargs)

    def __str__(self):
        return f'{self.name} - {self._func.__name__}'

    def _filter_returns_false(self, kwargs):
        return self._filter_by and not self._filter_by(kwargs)


class Callback:

    def __init__(self, subscription, suffix=None):
        self._subscription = subscription
        self._suffix = suffix

    def __call__(self, message):
        run_middleware_hook('pre_process_message', self._subscription)
        start_time = time.time()

        data = json.loads(message.data.decode('utf-8'))

        try:
            res = self._subscription(data, **dict(message.attributes))
        except Exception as e:
            run_middleware_hook('post_process_message_failure',
                                self._subscription, e, start_time)
        else:
            message.ack()
            run_middleware_hook('post_process_message_success',
                                self._subscription, start_time)
            return res
        finally:
            run_middleware_hook('post_process_message')


def sub(topic, prefix=None, suffix=None, filter_by=None):
    """Decorator function that makes declaring a PubSub Subscription simple.

    The Subscriber returned will automatically create and name
    the subscription for the topic.
    The subscription name will be the topic name prefixed by the project name.

    For example, if the topic name to subscribe too is `lets-tell-everyone`,
    the subscriber will be named `project-name-lets-tell-everyone`.

    Additionally, if a `suffix` param is added, the subscriber will be
    `project-name-lets-tell-everyone-my-suffix`.

    It is recommended to add `**kwargs` to your `sub` function. This will allow
    message attributes to be sent without breaking the subscriber
    implementation.

    Usage::

        @sub(topic='lets-tell-to-alice', prefix='shop')
        def bob_purpose(data, **kwargs):
             pass

        @sub(topic='lets-tell-everyone', suffix='sub1')
        def purpose_1(data, **kwargs):
             pass

        @sub(topic='lets-tell-everyone', suffix='sub2')
        def purpose_2(data, **kwargs):
             pass

        @sub(topic='photo-updated',
             filter_by=lambda **attrs: attrs.get('type') == 'landscape')
        def sub_process_landscape_photos(data, **kwargs):
            pass

    :param topic: string The topic that is being subscribed to.
    :param prefix: string An optional prefix to the subscription name.
                   Useful to namespace your subscription with your project name
    :param suffix: string An optional suffix to the subscription name.
                   Useful when you have two subscribers in the same project
                   that are subscribed to the same topic.
    :param filter_by: function An optional function that
                      filters the messages to be processed
                      by the sub regarding their attributes.
    :return: :class:`~rele.subscription.Subscription`
    """

    def decorator(func):
        return Subscription(func=func,
                            topic=topic,
                            prefix=prefix,
                            suffix=suffix,
                            filter_by=filter_by)

    return decorator



import json
import logging
import os
import time
from contextlib import suppress

from google.api_core import exceptions
from google.cloud import pubsub_v1
from rest_framework.utils import encoders

from rele.middleware import run_middleware_hook

logger = logging.getLogger(__name__)

USE_EMULATOR = True if os.environ.get('PUBSUB_EMULATOR_HOST') else False


class Subscriber:
    DEFAULT_ACK_DEADLINE = 60

    def __init__(self, gc_project_id, credentials):
        self._gc_project_id = gc_project_id
        if USE_EMULATOR:
            self._client = pubsub_v1.SubscriberClient()
        else:
            self._client = pubsub_v1.SubscriberClient(credentials=credentials)

    def get_default_ack_deadline(self):
        return int(os.environ.get('DEFAULT_ACK_DEADLINE',
                                  self.DEFAULT_ACK_DEADLINE))

    def create_subscription(
            self, subscription, topic, ack_deadline_seconds=None):
        ack_deadline_seconds = ack_deadline_seconds or \
            self.get_default_ack_deadline()

        subscription_path = self._client.subscription_path(
            self._gc_project_id, subscription)
        topic_path = self._client.topic_path(self._gc_project_id, topic)

        with suppress(exceptions.AlreadyExists):
            self._client.create_subscription(
                name=subscription_path,
                topic=topic_path,
                ack_deadline_seconds=ack_deadline_seconds)

    def consume(self, subscription_name, callback):
        subscription_path = self._client.subscription_path(
            self._gc_project_id, subscription_name)
        return self._client.subscribe(subscription_path, callback=callback)


class Publisher:
    """The Publisher Class

    Wraps the Google Cloud Publisher Client and handles.

    It is important that this class remains a Singleton class in the process.
    Otherwise, a memory leak will occur. To avoid this, it is strongly
    recommended to use the :meth:`~rele.publishing.publish` method.

    If the setting `USE_EMULATOR` evaluates to True, the Publisher Client will
    not have any credentials assigned.

    :param gc_project_id: string Google Cloud Project ID.
    :param credentials: string Google Cloud Credentials.
    :param timeout: integer, default 3.0 seconds.
    """
    def __init__(self, gc_project_id, credentials, timeout=3.0):
        self._gc_project_id = gc_project_id
        self._timeout = timeout
        if USE_EMULATOR:
            self._client = pubsub_v1.PublisherClient()
        else:
            self._client = pubsub_v1.PublisherClient(credentials=credentials)

    def publish(self, topic, data, blocking=False, **attrs):
        """Publishes message to Google PubSub topic.

        Usage::

            publisher = Publisher()
            publisher.publish('topic_name', {'foo': 'bar'})

        By default, this method is non-blocking, meaning that the method does
        not wait for the future to be returned.

        If you would like to wait for the future so you can track the message
        later, you can:

        Usage::

            publisher = Publisher()
            future = publisher.publish('topic_name', {'foo': 'bar'}, blocking=True) # noqa

        However, it should be noted that using `blocking=True` may incur a
        significant performance hit.

        In addition, the method adds a timestamp `published_at` to the
        message attrs using `epoch floating point number
        <https://docs.python.org/3/library/time.html#time.time>`_.

        :param topic: string topic to publish the data.
        :param data: dict with the content of the message.
        :param blocking: boolean, default False.
        :param attrs: Extra parameters to be published.
        :return: `Future <https://googleapis.github.io/google-cloud-python/latest/pubsub/subscriber/api/futures.html>`_  # noqa
        """

        attrs['published_at'] = str(time.time())
        run_middleware_hook('pre_publish', topic, data, attrs)
        payload = json.dumps(data, cls=encoders.JSONEncoder).encode('utf-8')
        topic_path = self._client.topic_path(self._gc_project_id, topic)
        future = self._client.publish(topic_path, payload, **attrs)
        if not blocking:
            return future

        future.result(timeout=self._timeout)
        run_middleware_hook('post_publish', topic)
        return future



import importlib

from rele.subscription import Subscription
from .middleware import register_middleware, default_middleware
from .publishing import init_global_publisher


class Config:

    def __init__(self, setting):
        self.gc_project_id = setting.get('GC_PROJECT_ID')
        self.credentials = setting.get('GC_CREDENTIALS')
        self.app_name = setting.get('APP_NAME')
        self.sub_prefix = setting.get('SUB_PREFIX')
        self.middleware = setting.get('MIDDLEWARE', default_middleware)


def setup(setting):
    config = Config(setting)
    init_global_publisher(config)
    register_middleware(config)
    return config


def load_subscriptions_from_paths(
        sub_module_paths, sub_prefix=None, filter_by=None):

    subscriptions = []
    for sub_module_path in sub_module_paths:
        sub_module = importlib.import_module(sub_module_path)
        for attr_name in dir(sub_module):
            attribute = getattr(sub_module, attr_name)
            if isinstance(attribute, Subscription):
                if sub_prefix and not attribute.prefix:
                    attribute.set_prefix(sub_prefix)

                if filter_by and not attribute.filter_by:
                    attribute.set_filter_by(filter_by)

                subscriptions.append(attribute)
    return subscriptions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

from setuptools import setup, find_packages


def get_version(*file_paths):
    """Retrieves the version from rele/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version('rele', '__init__.py')


if sys.argv[-1] == 'tag':
    print('Tagging the version on git: %s' % version)
    os.system('git tag -a %s -m "version %s"' % (version, version))
    os.system('git push --tags')
    sys.exit()

readme = open('README.md').read()

setup(
    name='rele',
    version=version,
    description="""Relé makes integration with Google PubSub easier.""",
    long_description=readme,
    long_description_content_type='text/markdown',
    author='MercadonaTech',
    author_email='software.online@mercadona.es',
    url='https://github.com/mercadona/rele',
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=['django', 'djangorestframework', 'google-cloud-pubsub'],
    license='Apache Software License 2.0',
    zip_safe=False,
    keywords='rele',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



__version__ = '0.4.1'
default_app_config = 'rele.apps.ReleConfig'

from .client import Publisher, Subscriber  # noqa
from .config import setup  # noqa
from .publishing import publish  # noqa
from .subscription import Callback, Subscription, sub  # noqa
from .worker import Worker  # noqa
